/*
 * Dark CSS theme for parce.
 *
 * Name "dark"
 */


/*
 * Defaults for a parce DIV/window etc.
 */

.parce {
    background: black;
    color: ivory;
    font: monospace 12pt;
}

.parce.current-line {
    background: #111;
}

.parce.current-line:focus {
    background: #181818;
}

.parce::selection {
    background: gray;
    color: white;
}

.parce::selection:focus {
    background: dodgerblue;
    color: white;
}

/* The current line when text widget is disabled
 * (probably unused normally)
 */
.parce.current-line:disabled {

}

.parce::selection:disabled {

}

/*
 * Styles for the base types
 */
.parce
.text {

}

.parce
.whitespace {

}

.parce
.keyword {
    font-weight: bold;
}

.parce
.delimiter {

}

.parce
.name {
    color: lightskyblue;
}

.parce
.literal {
    color: seagreen;
}

.parce
.comment {
    color: lightgray;
    font-family: serif;
    font-style: italic;
}

/*
 * Styles for the mixin types
 */
.parce
.alert {
    color: tomato;
    font-weight: bold;
}

.parce
.important {
    font-weight: bold;
}

.parce
.unimportant {

}

.parce
.special {
    color: forestgreen;
}

.parce
.definition {
    font-weight: bold;
}

.parce
.invalid {
    text-decoration: underline red solid;
}

.parce
.escape {
    font-weight: bold;
    color: darkcyan;
}

.parce
.template {
    font-weight: bold;
    color: green;
}

.parce
.pseudo {
    color: orange !important;
}

.parce
.preprocessed {
    color: green !important;
}

.parce
.inserted {

}

.parce
.deleted {

}

.parce
.quoted {

}

.parce
.bold {

}

.parce
.emphasized {

}

/* (Start, End, Indent and Dedent don't need styling) */
.parce
.start {

}

.parce
.end {

}

.parce
.indent {

}

.parce
.dedent {

}


/*
 * Styles for Name derivatives
 */
.parce
.attribute {
    color: dodgerblue;
}

.parce
.builtin {
    color: darksalmon;
    font-weight: bold;
}

.parce
.class {
    color: dodgerblue;
}

.parce
.command {

}

.parce
.constant {
    color: orchid;
    font-weight: bold;
}

.parce
.decorator {
    color: orange;
    font-weight: bold;
}

.parce
.entity {

}

.parce
.exception {
    color: crimson;
    font-weight: bold;
}

.parce
.function {

}

.parce
.identifier {
    color: brown;
}

.parce
.macro {

}

.parce
.markup {
    color: limegreen;
}

.parce
.method {

}

.parce
.namespace {

}

.parce
.object {
    color: darkorchid;
}

.parce
.property {
    color: royalblue;
}

.parce
.symbol {

}

.parce
.tag {
    font-weight: bold;
}

.parce
.type {

}

.parce
.variable {

}


/*
 * Styles for Literal derivatives
 */
.parce
.data {

}

.parce
.verbatim {

}

.parce
.string {
    color: indianred;
}

.parce
.number {
    color: pink;
}

.parce
.fraction {

}

.parce
.boolean {

}

.parce
.color {

}

.parce
.email {
    text-decoration: underline dodgerblue dashed;
}

.parce
.url {
    text-decoration: underline dodgerblue dashed;
}

.parce
.timestamp {

}

.parce
.input {

}

.parce
.output {

}

.parce
.error {

}

.parce
.prompt {

}

/*
 * String styles
 */
.parce
.string.double {

}

.parce
.string.single {

}

.parce
.string.escape {
    color: orangered;
}

/*
 * Styles for Delimiter derivatives
 */
.parce
.connection {

}

.parce
.direction {

}

.parce
.operator {

}

.parce
.assignment {

}

.parce
.bracket {
    font-weight: bold;
}

.parce
.modechange {
    font-weight: bold;
    color: chocolate;
}

.parce
.quote {

}

.parce
.separator {

}

