
.. Generated by ../update_languages.py -- do not edit!

.. list-table::
   :header-rows: 1
   :widths: 20 70

   * - Module
     - Languages

   * - :mod:`~parce.lang.css`
     - :class:`~parce.lang.css.Css`

   * - :mod:`~parce.lang.html`
     - :class:`~parce.lang.html.XHtml`, :class:`~parce.lang.html.Html`

   * - :mod:`~parce.lang.ini`
     - :class:`~parce.lang.ini.Ini`

   * - :mod:`~parce.lang.javascript`
     - :class:`~parce.lang.javascript.JavaScript`

   * - :mod:`~parce.lang.json`
     - :class:`~parce.lang.json.Json`

   * - :mod:`~parce.lang.lilypond`
     - :class:`~parce.lang.lilypond.LilyPond`

   * - :mod:`~parce.lang.python`
     - :class:`~parce.lang.python.Python`, :class:`~parce.lang.python.PythonConsole`

   * - :mod:`~parce.lang.scheme`
     - :class:`~parce.lang.scheme.Scheme`, :class:`~parce.lang.scheme.SchemeLily`

   * - :mod:`~parce.lang.tex`
     - :class:`~parce.lang.tex.Latex`

   * - :mod:`~parce.lang.texinfo`
     - :class:`~parce.lang.texinfo.Texinfo`

   * - :mod:`~parce.lang.toml`
     - :class:`~parce.lang.toml.Toml`

   * - :mod:`~parce.lang.troff`
     - :class:`~parce.lang.troff.Troff`

   * - :mod:`~parce.lang.xml`
     - :class:`~parce.lang.xml.Xml`, :class:`~parce.lang.xml.Dtd`

